/** @file
===========================================================================
MODULE				: MultiIOBaseFilter

DESCRIPTIONS		: This base class takes care of doing most of the menial work when writing a transform filter that has multiple input and output pins.

REVISION HISTORY	: Version 1.1
					  RG: Added EndOfStream, BeginFlush, EndFlush and NewSegment in case user needs to override extra methods

COPYRIGHT			: (c)CSIR, Meraka Institute 2007 all rights reserved

RESTRICTIONS		: The information/data/code contained within this file is 
					  the property of CSIR, Meraka Institute and has been 
					  classified as PUBLIC.
===========================================================================
*/

#include "MultiIOBaseFilter.h"

#include "MultiIOInputPin.h"
#include "MultiIOOutputPin.h"

#include <atlconv.h>

#pragma comment(lib, "atls.lib")

CMultiIOBaseFilter::CMultiIOBaseFilter(TCHAR *pObjectName, LPUNKNOWN lpUnk, CLSID clsid)
: CBaseFilter(pObjectName, lpUnk, &m_csFilter, clsid)
{}

CMultiIOBaseFilter::~CMultiIOBaseFilter(void)
{
	// Release all the COM interfaces
	for (int i = 0; i < m_vInputPins.size(); i++)
	{
		m_vInputPins[i]->Release();
	}
	for (int i = 0; i < m_vOutputPins.size(); i++)
	{
		m_vOutputPins[i]->Release();
	}

	// Clean up input type map
	CleanupTypeMap(m_mInputTypeMap);
	// Clean up Output type map
	CleanupTypeMap(m_mOutputTypeMap);
	// Clean up generic types
	for (int i = 0; i < m_vInputTypes.size(); i++)
	{
		delete m_vInputTypes.at(i);
	}
	for (int i = 0; i < m_vOutputTypes.size(); i++)
	{
		delete m_vOutputTypes.at(i);
	}
}

void CMultiIOBaseFilter::CleanupTypeMap(MEDIA_TYPE_MAP mTypeMap)
{
	MEDIA_TYPE_MAP_ITERATOR it;
	for (it = mTypeMap.begin(); it != mTypeMap.end(); it++)
	{
		//Delete the input type list for that pin
		MEDIA_TYPE_LIST* pInputList = (*it).second;
		//Iterate over list and delete all input types
		for (int i = 0; i < pInputList->size(); i++)
		{
			delete pInputList->at(i);
		}
		delete pInputList;
	}
	mTypeMap.clear();
}


void CMultiIOBaseFilter::Initialise()
{
	//Create initial number of inputs/outputs
	for (int i = 0; i < InitialNumberOfInputPins(); i++)
	{
		CreateInputPin();
	}

	for (int i = 0; i < InitialNumberOfOutputPins(); i++)
	{
		CreateOutputPin();
	}
	InitialiseInputTypes();
	InitialiseOutputTypes();
}

void CMultiIOBaseFilter::AddInputType( const GUID* pType, const GUID* pSubType, const GUID* pFormat, int nPin  )
{
	CIOMediaType* pMediaType = new CIOMediaType(pType, pSubType, pFormat);
	AddMediaType(pMediaType, m_vInputTypes, m_mInputTypeMap, nPin );
}

void CMultiIOBaseFilter::AddOutputType( const GUID* pType, const GUID* pSubType, const GUID* pFormat, int nPin   )
{
	CIOMediaType* pMediaType = new CIOMediaType(pType, pSubType, pFormat);
	AddMediaType(pMediaType, m_vOutputTypes, m_mOutputTypeMap, nPin );
}

void CMultiIOBaseFilter::AddMediaType(CIOMediaType* pMediaType, MEDIA_TYPE_LIST& typeList, MEDIA_TYPE_MAP& typeMap, int nPin  )
{
	if (nPin == -1)
	{
		// Generic types
		typeList.push_back(pMediaType);
	}
	else
	{
		// Add the mapping of the pin number to the type
		MEDIA_TYPE_MAP_ITERATOR it;
		it = typeMap.find(nPin);
		if (it == m_mInputTypeMap.end())
		{
			//No list for pin, create one
			typeMap[nPin] = new MEDIA_TYPE_LIST;
			typeMap[nPin]->push_back(pMediaType);
		}
		else
		{
			//list already exists for pin
			(*it).second->push_back(pMediaType);
		}
	}
}

int CMultiIOBaseFilter::GetPinCount()
{
	return (int)(m_vInputPins.size() + m_vOutputPins.size());
}

CBasePin * CMultiIOBaseFilter::GetPin( int n )
{
	if ((n >= GetPinCount()) || (n < 0))
		return NULL;

	if (n < (int) m_vInputPins.size())
	{
		//Return the input pin
		return m_vInputPins[n];
	}
	else
	{
		return m_vOutputPins[n - m_vInputPins.size() ];
	}
}

STDMETHODIMP CMultiIOBaseFilter::FindPin( LPCWSTR Id, IPin **ppPin )
{
	//Find pin according to it's name
	CheckPointer(ppPin,E_POINTER);
	ValidateReadWritePtr(ppPin,sizeof(IPin *));
	//Todo: Modify for this MUX: Check what ids this method is called with

	//Parse Id and get string number
	USES_CONVERSION;
	char* szInput = W2A(Id);
	char szTemp[10];
	ZeroMemory(szTemp, 10);
	memcpy(szTemp, szInput, 5);
	szTemp[5] = '\0';
	if (strcmp(szTemp, "Input")== 0)
	{
		char* szID = szInput + 6;
		int nId = atoi(szID);
		*ppPin = GetPin(nId);
	}
	else if (0==strcmp(szTemp, "Outpu")) 
	{
		char* szID = szInput + 7;
		int nId = atoi(szID);
		*ppPin = GetPin(m_vInputPins.size() + nId);
	} 
	else 
	{
		*ppPin = NULL;
		return VFW_E_NOT_FOUND;
	}

	HRESULT hr = NOERROR;
	//  AddRef() returned pointer - but GetPin could fail if memory is low.
	if (*ppPin) 
	{
		(*ppPin)->AddRef();
	} 
	else 
	{
		hr = E_OUTOFMEMORY;  // probably.  There's no pin anyway.
	}
	return hr;
}

void CMultiIOBaseFilter::CreateInputPin()
{
	HRESULT hr;
	USES_CONVERSION;
	char buffer[10] = "Input ";
	itoa(m_vInputPins.size(), buffer + 6, 10);
	CMultiIOInputPin* pInputPin = new CMultiIOInputPin(this, &m_csFilter, &hr, A2W(buffer), m_vInputPins.size());
	pInputPin->AddRef();
	m_vInputPins.push_back(pInputPin);
	// ensure enumerator is refreshed
	IncrementPinVersion();	
}

void CMultiIOBaseFilter::CreateOutputPin()
{
	HRESULT hr;
	USES_CONVERSION;
	char buffer[10] = "Output ";
	itoa(m_vOutputPins.size(), buffer + 7, 10);
	CMultiIOOutputPin* pOutputPin = new CMultiIOOutputPin(this, &m_csFilter, &hr, A2W(buffer), m_vOutputPins.size());
	pOutputPin->AddRef();
	m_vOutputPins.push_back(pOutputPin);
	// ensure enumerator is refreshed
	IncrementPinVersion();	
}

HRESULT CMultiIOBaseFilter::CheckInputType( const CMediaType *pMediaType )
{
	//RG 8/01/2008
	//Changing this to check all input types
	return FindMediaTypeInLists(pMediaType, &m_vInputTypes, m_mInputTypeMap);
}

HRESULT CMultiIOBaseFilter::FindMediaTypeInLists(const CMediaType* pMediaType, MEDIA_TYPE_LIST* pTypeList, MEDIA_TYPE_MAP& typeMap)
{
	// Search for media type in global list
	HRESULT hr = FindMediaType(pMediaType, pTypeList);
	if (SUCCEEDED(hr))
	{
		return hr;
	}

	//We haven't found it: Iterate over all pins own media types
	MEDIA_TYPE_MAP_ITERATOR it;
	for ( it = typeMap.begin(); it != typeMap.end(); it++)
	{
		MEDIA_TYPE_LIST* pTypeList = (*it).second;
		HRESULT hr = FindMediaType(pMediaType, pTypeList);
		if (SUCCEEDED(hr))
		{
			return hr;
		}
	}
	// We didn't find the type anywhere
	return VFW_E_TYPE_NOT_ACCEPTED;
}

void CMultiIOBaseFilter::OnConnect( int nIndex )
{
	// if this is the last one, make a new one: taken from Geraint's Mpeg Parser
	if (OnFullCreateMoreInputs())
	{
		if (nIndex == (int)(m_vInputPins.size()-1))
		{
			CreateInputPin();
		}
	}
}

void CMultiIOBaseFilter::OnDisconnect( int nIndex )
{
	if (OnFullCreateMoreInputs())
	{
		// if index is the last but one, and
		// the last one is still unconnected, then
		// remove the last one
		if ((nIndex == (int)(m_vInputPins.size() - 2)) &&
			(!m_vInputPins[m_vInputPins.size()-1]->IsConnected()))
		{
			// Now make sure that we can't disconnect more than the initial number of inputs
			// You cannot remove a pin to have less pins than you started with
			if (m_vInputPins.size() > InitialNumberOfInputPins())
			{
				m_vInputPins[m_vInputPins.size()-1]->Release();
				m_vInputPins.pop_back();
				// refresh enumerator
				IncrementPinVersion();
			}
		}
	}
}


int CMultiIOBaseFilter::InitialNumberOfInputPins()
{
	return 1;
}

int CMultiIOBaseFilter::InitialNumberOfOutputPins()
{
	return 1;
}

bool CMultiIOBaseFilter::OnFullCreateMoreInputs()
{
	return false;
}

bool CMultiIOBaseFilter::OnFullCreateMoreOutputs()
{
	return false;
}

//////////////////////////////////////////////////////////////////////////
// Output pin stuff
HRESULT CMultiIOBaseFilter::CheckOutputType( const CMediaType* pMediaType )
{
	return FindMediaTypeInLists(pMediaType, &m_vOutputTypes, m_mOutputTypeMap);
}

/**
* 
* 
* Method:    InitializeOutputSample
* FullName:  CMultiIOBaseFilter::InitializeOutputSample
* Access:    virtual public 
* Qualifier:
* @return:   HRESULT
* @param: 	 IMediaSample *pSample
* @param: 	 IMediaSample **ppOutSample
* @param: 	 int nInputIndex: this is required when initialising an out sample so that we can get the properties such as sample start time of the specified input and use this for the specified out sample
* @param: 	 int nOutputIndex: this needs to be specified so that we can get the correct allocator
*/
HRESULT CMultiIOBaseFilter::InitializeOutputSample( IMediaSample *pSample, IMediaSample **ppOutSample, int nInputIndex, int nOutputIndex )
{
	AM_SAMPLE2_PROPERTIES * const pProperties = m_vInputPins[nInputIndex]->SampleProps();

	IMediaSample *pOutSample;
	//Let's get rid of the const

	AM_SAMPLE2_PROPERTIES* pProps = (AM_SAMPLE2_PROPERTIES*)pProperties;
	// default - times are the same
	DWORD dwFlags = 0;
	// This will prevent the image renderer from switching us to DirectDraw
	// when we can't do it without skipping frames because we're not on a
	// keyframe.  If it really has to switch us, it still will, but then we
	// will have to wait for the next keyframe
	if (!(pProps->dwSampleFlags & AM_SAMPLE_SPLICEPOINT)) {
		dwFlags |= AM_GBF_NOTASYNCPOINT;
	}

	ASSERT(m_vOutputPins[nOutputIndex]->GetAllocator() != NULL);
	HRESULT hr = m_vOutputPins[nOutputIndex]->GetAllocator()->GetBuffer(
		&pOutSample
		, pProps->dwSampleFlags & AM_SAMPLE_TIMEVALID ?
		&pProps->tStart : NULL
		, pProps->dwSampleFlags & AM_SAMPLE_STOPVALID ?
		&pProps->tStop : NULL
		, dwFlags
		);
	*ppOutSample = pOutSample;
	if (FAILED(hr)) {
		return hr;
	}

	ASSERT(pOutSample);
	IMediaSample2 *pOutSample2;
	if (SUCCEEDED(pOutSample->QueryInterface(IID_IMediaSample2,
		(void **)&pOutSample2))) {
			/*  Modify it */
			AM_SAMPLE2_PROPERTIES OutProps;
			EXECUTE_ASSERT(SUCCEEDED(pOutSample2->GetProperties(
				FIELD_OFFSET(AM_SAMPLE2_PROPERTIES, tStart), (PBYTE)&OutProps)
				));
			OutProps.dwTypeSpecificFlags = pProps->dwTypeSpecificFlags;
			OutProps.dwSampleFlags =
				(OutProps.dwSampleFlags & AM_SAMPLE_TYPECHANGED) |
				(pProps->dwSampleFlags & ~AM_SAMPLE_TYPECHANGED);
			OutProps.tStart = pProps->tStart;
			OutProps.tStop  = pProps->tStop;
			OutProps.cbData = FIELD_OFFSET(AM_SAMPLE2_PROPERTIES, dwStreamId);
			hr = pOutSample2->SetProperties(
				FIELD_OFFSET(AM_SAMPLE2_PROPERTIES, dwStreamId),
				(PBYTE)&OutProps
				);
			pOutSample2->Release();
	} else {
		if (pProps->dwSampleFlags & AM_SAMPLE_TIMEVALID) {
			pOutSample->SetTime(&pProps->tStart,
				&pProps->tStop);
		}
		if (pProps->dwSampleFlags & AM_SAMPLE_SPLICEPOINT) {
			pOutSample->SetSyncPoint(TRUE);
		}
		if (pProps->dwSampleFlags & AM_SAMPLE_DATADISCONTINUITY) {
			pOutSample->SetDiscontinuity(TRUE);
		}
		// Copy the media times

		LONGLONG MediaStart, MediaEnd;
		if (pSample->GetMediaTime(&MediaStart,&MediaEnd) == NOERROR) {
			pOutSample->SetMediaTime(&MediaStart,&MediaEnd);
		}
	}
	return S_OK;
}

HRESULT CMultiIOBaseFilter::SetMediaType( PIN_DIRECTION direction,const CMediaType *pmt, int nIndex )
{
	UNREFERENCED_PARAMETER(direction);
	UNREFERENCED_PARAMETER(pmt);
	//Grip a copy of the media type
	if (direction == PINDIR_INPUT)
	{
		m_mMediaTypes[nIndex] = *pmt;
	}
	// Check if pin at position accepts the input type
	//First iterate over all common media types
	HRESULT hr = FindMediaType(pmt, &m_vInputTypes);
	if (SUCCEEDED(hr))
	{
		return hr;
	}

	//We haven't found it: Iterate over pins own media types
	MEDIA_TYPE_MAP_ITERATOR it;
	it = m_mInputTypeMap.find(nIndex);
	if (it == m_mInputTypeMap.end())
	{
		//We couldn't find it, reject the connection
		return E_FAIL;
	}
	else
	{
		MEDIA_TYPE_LIST* pInputTypeList = (*it).second;
		HRESULT hr = FindMediaType(pmt, pInputTypeList);
		return hr;
	}
}

HRESULT CMultiIOBaseFilter::FindMediaType(const CMediaType* pmt, MEDIA_TYPE_LIST* pInputTypeList)
{
	for (int i = 0; i < pInputTypeList->size(); i++)
	{
		//Compare major type
		const GUID* pType = (pInputTypeList->at(i))->MediaType;
		if (((pmt->majortype == *pType) && (pType))||(!pType))
		{
			//Compare sub type
			const GUID* pSubtype = (pInputTypeList->at(i))->MediaSubtype;
			if (((pmt->subtype == *pSubtype) && (pSubtype))||(!pSubtype))
			{
				//Compare formats
				const GUID* pFormat = (pInputTypeList->at(i))->Format;
				if (((pmt->formattype == *pFormat)&&(pFormat))||(!pFormat))
				{
					return NOERROR;
				}
			}
		}
	}
	return E_FAIL;
}

// Basic implementation -> Just here for override purposes
STDMETHODIMP CMultiIOBaseFilter::EndOfStream( int nIndex )
{
	return S_OK;
}

STDMETHODIMP CMultiIOBaseFilter::BeginFlush( int nIndex )
{
	return S_OK;
}

STDMETHODIMP CMultiIOBaseFilter::EndFlush( int nIndex )
{
	return S_OK;
}

STDMETHODIMP CMultiIOBaseFilter::NewSegment( REFERENCE_TIME tStart, REFERENCE_TIME tStop, double dRate, int nIndex )
{
	return S_OK;
}