/** @file
===========================================================================
MODULE				: CCropFilter

DESCRIPTIONS		: Crop Filter for RGB24 and RGB32 Media

REVISION HISTORY	: Version 1.0
					27/08/2008: RG Fixed boolean logic bug in CheckTransform which resulted in the VMR9 rendering an incorrect image
					27/08/2008: RG Adjusted source and target rect in GetMediaType
					01/09/2008: RG Changed interface: crop parameters are now specified as crop amounts and NOT as ratios.


COPYRIGHT			: (c)CSIR, Meraka Institute 2007 all rights reserved

RESTRICTIONS		: The information/data/code contained within this file is 
					  the property of CSIR, Meraka Institute and has been 
					  classified as PUBLIC.
===========================================================================
*/
#include "CropFilter.h"

#include <DirectShow/Common/datastructures.h>

#include <Utils/ImageUtils/PicCropperRGB24Impl.h>
#include <Utils/ImageUtils/PicCropperRGB32Impl.h>

CCropFilter::CCropFilter()
: CCustomBaseFilter(NAME("Meraka RTVC Crop Filter"), 0, CLSID_CropFilter),
	m_pCropper(NULL),
	m_nBytesPerPixel(BYTES_PER_PIXEL_RGB24),
	m_nLeftCrop(0),
	m_nRightCrop(0),
	m_nTopCrop(0),
	m_nBottomCrop(0)
{
	//Call the initialise input method to load all acceptable input types for this filter
	InitialiseInputTypes();
	//Default values
	m_nOutWidth = 0;
	m_nOutHeight = 0;
}

CCropFilter::~CCropFilter()
{
	if (m_pCropper)
	{
		delete m_pCropper;
		m_pCropper = NULL;
	}
}

CUnknown * WINAPI CCropFilter::CreateInstance( LPUNKNOWN pUnk, HRESULT *pHr )
{
	CCropFilter *pFilter = new CCropFilter();
	if (pFilter== NULL) 
	{
		*pHr = E_OUTOFMEMORY;
	}
	return pFilter;
}


void CCropFilter::InitialiseInputTypes()
{
	AddInputType(&MEDIATYPE_Video, &MEDIASUBTYPE_RGB24, &FORMAT_VideoInfo);
	AddInputType(&MEDIATYPE_Video, &MEDIASUBTYPE_RGB32, &FORMAT_VideoInfo);
}

HRESULT CCropFilter::SetMediaType( PIN_DIRECTION direction, const CMediaType *pmt )
{
	HRESULT hr = CCustomBaseFilter::SetMediaType(direction, pmt);
	if (direction == PINDIR_INPUT)
	{
		//Set defaults and make sure that the target dimensions are valid
		if ((m_nOutWidth == 0)||(m_nOutWidth > m_nInWidth))
		{
			m_nOutWidth = m_nInWidth;
		}
		if ((m_nOutHeight == 0)||(m_nOutHeight > m_nInHeight))
		{
			m_nOutHeight = m_nInHeight;
		}
		//Determine whether we are connected to a RGB24 or 32 source
		if (pmt->majortype == MEDIATYPE_Video)
		{
			//The converter might already exist if the filter has been connected previously
			if (m_pCropper)
			{
				delete m_pCropper;
				m_pCropper = NULL;
			}
			if (pmt->subtype == MEDIASUBTYPE_RGB24)
			{
				m_pCropper = new PicCropperRGB24Impl();
				m_nBytesPerPixel = BYTES_PER_PIXEL_RGB24;
			}
			else if (pmt->subtype == MEDIASUBTYPE_RGB32)
			{
				m_pCropper = new PicCropperRGB32Impl();
				m_nBytesPerPixel = BYTES_PER_PIXEL_RGB32;
			}
		}
		RecalculateFilterParameters();
	}
	return hr;
}

HRESULT CCropFilter::GetMediaType( int iPosition, CMediaType *pMediaType )
{
	if (iPosition < 0)
	{
		return E_INVALIDARG;
	}
	else if (iPosition == 0)
	{
		// Get the input pin's media type and return this as the output media type - we want to retain
		// all the information about the image
		HRESULT hr = m_pInput->ConnectionMediaType(pMediaType);
		if (FAILED(hr))
		{
			return hr;
		}
		// Get the bitmap info header and adapt the cropped 
		//make sure that it's a video info header
		ASSERT(pMediaType->formattype == FORMAT_VideoInfo);
		VIDEOINFOHEADER *pVih = (VIDEOINFOHEADER*)pMediaType->pbFormat;
		//Now we need to calculate the size of the output image
		BITMAPINFOHEADER* pBi = &(pVih->bmiHeader);
		
		pBi->biHeight = m_nOutHeight;
		ASSERT(pBi->biHeight > 0);

		pBi->biWidth = m_nOutWidth;
		ASSERT(pBi->biWidth > 0);
		pBi->biSizeImage = pBi->biWidth * pBi->biHeight * m_nBytesPerPixel;

		pVih->rcSource.top = 0;
		pVih->rcSource.left = 0;
		pVih->rcSource.right = pBi->biWidth;
		pVih->rcSource.bottom = pBi->biHeight;

		pVih->rcTarget.top = 0;
		pVih->rcTarget.left = 0;
		pVih->rcTarget.right = pBi->biWidth;
		pVih->rcTarget.bottom = pBi->biHeight;

		pMediaType->lSampleSize = pBi->biSizeImage;

		return S_OK;
	}
	return VFW_S_NO_MORE_ITEMS;
}

HRESULT CCropFilter::DecideBufferSize( IMemAllocator *pAlloc, ALLOCATOR_PROPERTIES *pProp )
{
	pProp->cbBuffer = m_nOutPixels * m_nBytesPerPixel;

	if (pProp->cbAlign == 0)
	{
		pProp->cbAlign = 1;
	}
	if (pProp->cBuffers == 0)
	{
		pProp->cBuffers = 1;
	}
	
	// Set allocator properties.
	ALLOCATOR_PROPERTIES Actual;
	HRESULT hr = pAlloc->SetProperties(pProp, &Actual);
	if (FAILED(hr)) 
	{
		return hr;
	}
	// Even when it succeeds, check the actual result.
	if (pProp->cbBuffer > Actual.cbBuffer) 
	{
		return E_FAIL;
	}
	return S_OK;
}

HRESULT CCropFilter::CheckTransform( const CMediaType *mtIn, const CMediaType *mtOut )
{
	//Make sure the input and output types are related
	if (mtOut->majortype != MEDIATYPE_Video)
	{
		return VFW_E_TYPE_NOT_ACCEPTED;
	}
	// RG: 27/08/2008 Bug FIX: Resulted in incorrect image in VMR9
	if (!
		((mtIn->subtype == MEDIASUBTYPE_RGB24)&&(mtOut->subtype == MEDIASUBTYPE_RGB24))
		||
		((mtIn->subtype == MEDIASUBTYPE_RGB32)&&(mtOut->subtype == MEDIASUBTYPE_RGB32))
		)
	{
		return VFW_E_TYPE_NOT_ACCEPTED;
	}
	if (mtOut->formattype != FORMAT_VideoInfo)
	{
		return VFW_E_TYPE_NOT_ACCEPTED;
	}
	return S_OK;
}


STDMETHODIMP CCropFilter::GetParameter( const char* type, char* szValue, int& nResultLength )
{
	char *p = (char *)type;
	int len	= (int)strlen(p);

	if( _strnicmp(p, TARGET_WIDTH,len) == 0 )
		_itoa(m_nOutWidth,(char *)szValue,10);
	else if( _strnicmp(p, TARGET_HEIGHT, len) == 0 )
		_itoa(m_nOutHeight,(char *)szValue,10);
	else if( _strnicmp(p, LEFT_CROP, len) == 0 )
		_itoa(m_nLeftCrop,(char *)szValue,10);
	else if( _strnicmp(p, RIGHT_CROP, len) == 0 )
		_itoa(m_nRightCrop,(char *)szValue,10);
	else if( _strnicmp(p, TOP_CROP, len) == 0 )
		_itoa(m_nTopCrop,(char *)szValue,10);
	else if( _strnicmp(p, BOTTOM_CROP, len) == 0 )
		_itoa(m_nBottomCrop,(char *)szValue,10);
	else
	{
		//_errorStr = "Cropper: Read parameter not supported";
		nResultLength = -1;
		return S_FALSE;
	}//end else...

	nResultLength = (int)strlen(szValue);
	return S_OK;
}

STDMETHODIMP CCropFilter::SetParameter( const char* type, const char* value )
{
	// For now, one cannot set any parameters once the output has been connected -> this will affect the buffer size
	// m_pOutput only gets created in the Trnasform filter once the input pin has been connected - need to check pointer
	if (m_pOutput)
	{
		if (m_pOutput->IsConnected())
		{
			DbgLog((LOG_TRACE, 0, TEXT("Unable to set parameter: output pins is already connected - no dynamic reconnection support")));
			return E_FAIL;
		}
	}

	char *p = (char *)type;
	char *v = (char *)value;
	int  len	= (int)strlen(p);

	if( _strnicmp(p, LEFT_CROP, len) == 0 )
	{
		int nCrop = (atoi(v));
		return SetCropIfValid(m_nInWidth, nCrop, m_nLeftCrop, m_nRightCrop);
	}
	else if( _strnicmp(p, RIGHT_CROP, len) == 0 )
	{
		int nCrop = (atoi(v));
		return SetCropIfValid(m_nInWidth, nCrop, m_nRightCrop, m_nLeftCrop);
	}
	else if( _strnicmp(p, TOP_CROP, len) == 0 )
	{
		int nCrop = (atoi(v));
		return SetCropIfValid(m_nInHeight, nCrop, m_nTopCrop, m_nBottomCrop);
	}
	else if( _strnicmp(p, BOTTOM_CROP, len) == 0 )
	{
		int nCrop = (atoi(v));
		return SetCropIfValid(m_nInHeight, nCrop, m_nBottomCrop, m_nTopCrop);
	}
	else
	{
		//_errorStr = "Cropper: Write parameter not supported";
		return E_FAIL;
	}//end else...
}

DWORD CCropFilter::ApplyTransform( BYTE* pBufferIn, BYTE* pBufferOut )
{
	int nTotalSize = 0;
	//make sure we were able to initialise our converter
	if (m_pCropper)
	{
		//Call cropping conversion code
		m_pCropper->SetInDimensions(m_nInWidth, m_nInHeight);
		m_pCropper->SetOutDimensions(m_nOutWidth, m_nOutHeight);
		m_pCropper->SetCrop(m_nLeftCrop, m_nRightCrop, m_nTopCrop, m_nBottomCrop);
		m_pCropper->Crop((void*)pBufferIn, (void*)pBufferOut);
		//Total pixels is now 4 times larger
		nTotalSize = m_nOutWidth * m_nOutHeight * m_nBytesPerPixel;
	}
	else
	{
		DbgLog((LOG_TRACE, 0, TEXT("TestCropper: Cropper is not initialised - unable to transform")));
	}
	return nTotalSize;
}

void CCropFilter::RecalculateFilterParameters()
{
	//Set defaults and make sure that the target dimensions are valid
	m_nOutWidth = m_nInWidth - m_nLeftCrop - m_nRightCrop;
	m_nOutHeight = m_nInHeight - m_nTopCrop - m_nBottomCrop;
	m_nOutPixels = m_nOutWidth * m_nOutHeight;
}

HRESULT CCropFilter::SetCropIfValid( int nTotalDimensionImage, int nNewCrop, int& nOldCrop, int nOppositeCrop  )
{
	if (nNewCrop >= 0)
	{
		if (nTotalDimensionImage - nOppositeCrop - nNewCrop > 0)
		{
			nOldCrop = nNewCrop;
			RecalculateFilterParameters();
			return S_OK;
		}
		else
		{
			// Fail: Total crop exceeds image dimension
			return E_FAIL;
		}
	}
	else
	{
		// negative value
		return E_FAIL;
	}
}